! WHIZARD 2.2.0_alpha-4 Nov 15 2013
! 
! Copyright (C) 1999-2013 by 
!     Wolfgang Kilian <kilian@physik.uni-siegen.de>
!     Thorsten Ohl <ohl@physik.uni-wuerzburg.de>
!     Juergen Reuter <juergen.reuter@desy.de>
!     
!     with contributions by 
!     Christian Speckner <cnspeckn@googlemail.com> 
!     and  Felix Braam, Sebastian Schmidt, Daniel Wiesler 
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module decays

  use kinds, only: default !NODEP!
  use iso_varying_string, string_t => varying_string !NODEP!
  use file_utils !NODEP!
  use diagnostics !NODEP!
  use unit_tests
  use os_interface

  use flavors
  use interactions
  use variables
  use models
  use selectors
  use prc_core
  use prc_test
  use process_libraries
  use mci_base
  use mci_midpoint
  use phs_base
  use phs_single
  use processes
  
  implicit none
  private

  public :: decays_test

  type, abstract :: any_config_t
   contains
     procedure (particle_config_write), deferred :: write
  end type any_config_t

  type :: particle_config_t
     class(any_config_t), allocatable :: c
  end type particle_config_t
  
  type :: decay_term_config_t
     type(particle_config_t), dimension(:), allocatable :: prt
   contains
     procedure :: write => decay_term_config_write
     procedure :: init => decay_term_config_init
  end type decay_term_config_t
  
  type :: decay_term_t
     type(decay_term_config_t), pointer :: config => null ()
     type(unstable_t), dimension(:), pointer :: unstable_product => null ()
   contains
     procedure :: final => decay_term_final
     procedure :: write => decay_term_write
     procedure :: write_process_instances => decay_term_write_process_instances
     procedure :: init => decay_term_init
  end type decay_term_t

  type :: decay_config_t
     type(string_t) :: process_id
     type(process_t), pointer :: process => null ()
     real(default) :: weight = 0
     real(default) :: integral = 0
     real(default) :: abs_error = 0
     real(default) :: rel_error = 0
     type(decay_term_config_t), dimension(:), allocatable :: term_config
   contains
     procedure :: final => decay_config_final
     procedure :: write => decay_config_write
     procedure :: init => decay_config_init
     procedure :: init_term => decay_config_init_term
     procedure :: connect => decay_connect
     procedure :: compute => decay_config_compute
  end type decay_config_t
  
  type :: decay_t
     type(decay_config_t), pointer :: config => null ()
     integer :: selected = 0
     type(decay_term_t), dimension(:), allocatable :: term
     type(process_instance_t), pointer :: process_instance => null ()
   contains
     procedure :: final => decay_final
     procedure :: write => decay_write
     procedure :: write_process_instances => decay_write_process_instances
     procedure :: init => decay_init
  end type decay_t
     
  type, extends (any_config_t) :: stable_config_t
     type(flavor_t), dimension(:), allocatable :: flv
   contains
     procedure :: write => stable_config_write
     procedure :: init => stable_config_init
  end type stable_config_t

  type, extends (any_config_t) :: unstable_config_t
     type(flavor_t) :: flv
     real(default) :: integral = 0
     real(default) :: abs_error = 0
     real(default) :: rel_error = 0
     type(selector_t) :: selector
     type(decay_config_t), dimension(:), allocatable :: decay_config
   contains
     procedure :: final => unstable_config_final
     procedure :: write => unstable_config_write
     procedure :: init => unstable_config_init
     procedure :: init_decays => unstable_config_init_decays
     procedure :: compute => unstable_config_compute
  end type unstable_config_t
  
  type :: unstable_t
     type(unstable_config_t), pointer :: config => null ()
     integer :: mode = 0
     type(decay_t), dimension(:), allocatable :: decay
   contains
     procedure :: final => unstable_final
     procedure :: write => unstable_write
     procedure :: write_process_instances => unstable_write_process_instances
     procedure :: init => unstable_init
  end type unstable_t
  

  interface
     subroutine particle_config_write (object, unit, indent, verbose)
       import
       class(any_config_t), intent(in) :: object
       integer, intent(in), optional :: unit, indent
       logical, intent(in), optional :: verbose
     end subroutine particle_config_write
  end interface


contains

  recursive subroutine decay_term_config_write (object, unit, indent, verbose)
    class(decay_term_config_t), intent(in) :: object
    integer, intent(in), optional :: unit, indent
    logical, intent(in), optional :: verbose
    integer :: i, j, u, ind
    logical :: verb
    u = output_unit (unit)
    ind = 0;  if (present (indent))  ind = indent
    verb = .true.;  if (present (verbose))  verb = verbose
    call write_indent (u, ind)
    write (u, "(A)", advance="no")  "Final state:"
    do i = 1, size (object%prt)
       select type (prt_config => object%prt(i)%c)
       type is (stable_config_t)
          write (u, "(1x,A)", advance="no") &
               char (flavor_get_name (prt_config%flv(1)))
          do j = 2, size (prt_config%flv)
             write (u, "(':',A)", advance="no") &
                  char (flavor_get_name (prt_config%flv(j)))
          end do
       type is (unstable_config_t)
          write (u, "(1x,A)", advance="no") &
               char (flavor_get_name (prt_config%flv))
       end select
    end do
    write (u, *)
    if (verb) then
       do i = 1, size (object%prt)
          call object%prt(i)%c%write (u, ind)
       end do
    end if
  end subroutine decay_term_config_write

  subroutine decay_term_config_init (term, flv, stable)
    class(decay_term_config_t), intent(out) :: term
    type(flavor_t), dimension(:,:), intent(in) :: flv
    logical, dimension(:), intent(in) :: stable
    integer :: i
    allocate (term%prt (size (flv, 1)))
    do i = 1, size (flv, 1)
       associate (prt => term%prt(i))
         if (stable(i)) then
            allocate (stable_config_t :: prt%c)
         else
            allocate (unstable_config_t :: prt%c)
         end if
         select type (prt_config => prt%c)
         type is (stable_config_t)
            call prt_config%init (flv(i,:))
         type is (unstable_config_t)
            if (all (flv(i,:) == flv(i,1))) then
               call prt_config%init (flv(i,1))
            else
               call prt_config%write ()
               call msg_fatal ("Decay configuration: &
                    &unstable product must be unique")
            end if
         end select
       end associate
    end do
  end subroutine decay_term_config_init
  
  recursive subroutine decay_term_final (object)
    class(decay_term_t), intent(inout) :: object
    integer :: i
    if (associated (object%unstable_product)) then
       do i = 1, size (object%unstable_product)
          call object%unstable_product(i)%final ()
       end do
       deallocate (object%unstable_product)
    end if
  end subroutine decay_term_final
  
  recursive subroutine decay_term_write (object, unit, indent)
    class(decay_term_t), intent(in) :: object
    integer, intent(in), optional :: unit, indent
    integer :: i, j, u, ind
    u = output_unit (unit)
    ind = 0;  if (present (indent))  ind = indent
    call object%config%write (u, ind, verbose = .false.)
    j = 0
    do i = 1, size (object%config%prt)
       select type (prt_config => object%config%prt(i)%c)
       type is (stable_config_t)
          call prt_config%write (u, ind)
       type is (unstable_config_t)
          j = j + 1
          call object%unstable_product(j)%write (u, ind)
       end select
    end do
  end subroutine decay_term_write
    
  subroutine decay_term_write_process_instances (term, unit)
    class(decay_term_t), intent(in) :: term
    integer, intent(in), optional :: unit
    integer :: i
    do i = 1, size (term%unstable_product)
       call term%unstable_product(i)%write_process_instances (unit)
    end do
  end subroutine decay_term_write_process_instances
  
  subroutine decay_term_init (term, config)
    class(decay_term_t), intent(out) :: term
    type(decay_term_config_t), intent(in), target :: config
    integer :: n_unstable, i
    term%config => config
    n_unstable = 0
    do i = 1, size (config%prt)
       select type (prt_config => config%prt(i)%c)
       type is (unstable_config_t);  n_unstable = n_unstable + 1
       end select
    end do
    allocate (term%unstable_product (n_unstable))
  end subroutine decay_term_init

  recursive subroutine decay_config_final (object)
    class(decay_config_t), intent(inout) :: object
  end subroutine decay_config_final
  
  recursive subroutine decay_config_write (object, unit, indent, verbose)
    class(decay_config_t), intent(in) :: object
    integer, intent(in), optional :: unit, indent
    logical, intent(in), optional :: verbose
    integer :: i, u, ind
    logical :: verb
    u = output_unit (unit)
    ind = 0;  if (present (indent))  ind = indent
    verb = .true.;  if (present (verbose))  verb = verbose
    call write_indent (u, ind)
    write (u, "(A)")  "Decay:"
    call write_indent (u, ind)
    if (associated (object%process)) then
       write (u, 3)  "process ID      =", char (object%process_id), "*"
    else
       write (u, 3)  "process ID      =", char (object%process_id)
    end if
    call write_indent (u, ind)
    write (u, 2)  "branching ratio =", object%weight * 100
    call write_indent (u, ind)
    write (u, 1)  "partial width   =", object%integral
    call write_indent (u, ind)
    write (u, 1)  "error (abs)     =", object%abs_error
    call write_indent (u, ind)
    write (u, 1)  "error (rel)     =", object%rel_error
1   format (2x,A,ES19.12)
2   format (2x,A,F11.6,1x,'%')
3   format (2x,A,2(1x,A))
    if (verb .and. allocated (object%term_config)) then
       do i = 1, size (object%term_config)
          call object%term_config(i)%write (u, ind + 1)
       end do
    end if
  end subroutine decay_config_write
    
  subroutine decay_config_init (decay, process_id, n_terms)
    class(decay_config_t), intent(out) :: decay
    type(string_t), intent(in) :: process_id
    integer, intent(in), optional :: n_terms
    decay%process_id = process_id
    if (present (n_terms)) then
       allocate (decay%term_config (n_terms))
    end if
  end subroutine decay_config_init
       
  subroutine decay_config_init_term (decay, i, flv, stable)
    class(decay_config_t), intent(inout) :: decay
    integer, intent(in) :: i
    type(flavor_t), dimension(:,:), intent(in) :: flv
    logical, dimension(:), intent(in) :: stable
    call decay%term_config(i)%init (flv, stable)
  end subroutine decay_config_init_term
  
  subroutine decay_connect (decay, process)
    class(decay_config_t), intent(out) :: decay
    type(process_t), intent(in), target :: process
    type(flavor_t), dimension(:,:), allocatable :: flv
    logical, dimension(:), allocatable :: stable
    integer :: i, j
    call decay%init (process%get_id (), process%get_n_terms ())
    do i = 1, size (decay%term_config)
       call process%get_term_flavor (i, flv)
       allocate (stable (size (flv, 1) - 1))
       stable = flavor_is_stable (flv(2:,1))
       call decay%init_term (i, flv(2:,:), stable)
       deallocate (flv)
    end do
    decay%integral = process%get_integral ()
    decay%abs_error = process%get_error ()
    decay%process => process
  end subroutine decay_connect

  recursive subroutine decay_config_compute (decay)
    class(decay_config_t), intent(inout) :: decay
    if (decay%integral == 0) then
       call decay%write ()
       call msg_fatal ("Decay configuration: partial width is zero")
    end if
    decay%rel_error = decay%abs_error / decay%integral
  end subroutine decay_config_compute
  
  recursive subroutine decay_final (object)
    class(decay_t), intent(inout) :: object
    integer :: i
    if (allocated (object%term)) then
       do i = 1, size (object%term)
          call object%term(i)%final ()
       end do
    end if
    call object%process_instance%final ()
    deallocate (object%process_instance)
  end subroutine decay_final
  
  recursive subroutine decay_write (object, unit, indent)
    class(decay_t), intent(in) :: object
    integer, intent(in), optional :: unit, indent
    integer :: i, u, ind
    u = output_unit (unit)
    ind = 0;  if (present (indent))  ind = indent
    call object%config%write (unit, indent, verbose = .false.)
    call write_indent (u, ind)
    if (object%selected > 0) then
       write (u, "(2x,A,I0)")  "Selected term   = ", object%selected
       call object%term(object%selected)%write (u, ind + 1)
    else
       write (u, "(2x,A)")  "Selected term   = [undefined]"
    end if
  end subroutine decay_write

  recursive subroutine decay_write_process_instances (decay, unit)
    class(decay_t), intent(in) :: decay
    integer, intent(in), optional :: unit
    if (associated (decay%process_instance)) then
       call decay%process_instance%write (unit)
    end if
    if (decay%selected > 0) then
       call decay%term(decay%selected)%write_process_instances (unit)
    end if
  end subroutine decay_write_process_instances
    
  recursive subroutine decay_init (decay, config)
    class(decay_t), intent(out) :: decay
    type(decay_config_t), intent(in), target :: config
    integer :: i
    decay%config => config
    allocate (decay%process_instance)
    call decay%process_instance%init (decay%config%process)
    call decay%process_instance%setup_event_data ()
    call decay%config%process%prepare_simulation (1)
    call decay%process_instance%init_simulation (1)
    allocate (decay%term (size (config%term_config)))
    do i = 1, size (config%term_config)
       call decay%term(i)%init (config%term_config(i))
    end do
  end subroutine decay_init

  recursive subroutine stable_config_write (object, unit, indent, verbose)
    class(stable_config_t), intent(in) :: object
    integer, intent(in), optional :: unit, indent
    logical, intent(in), optional :: verbose
    integer :: u, i, ind
    u = output_unit (unit)
    ind = 0;  if (present (indent))  ind = indent
    call write_indent (u, ind)
    write (u, "('+',1x,A)", advance = "no")  "Stable:"
    write (u, "(1x,A)", advance = "no")  char (flavor_get_name (object%flv(1)))
    do i = 2, size (object%flv)
       write (u, "(':',A)", advance = "no") &
            char (flavor_get_name (object%flv(i)))
    end do
    write (u, *)
  end subroutine stable_config_write
  
  subroutine stable_config_init (config, flv)
    class(stable_config_t), intent(out) :: config
    type(flavor_t), dimension(:), intent(in) :: flv
    integer, dimension (size (flv)) :: pdg
    logical, dimension (size (flv)) :: mask
    integer :: i
    pdg = flavor_get_pdg (flv)
    mask(1) = .true.
    forall (i = 2 : size (pdg))
       mask(i) = all (pdg(i) /= pdg(1:i-1))
    end forall
    allocate (config%flv (count (mask)))
    config%flv = pack (flv, mask)
  end subroutine stable_config_init
  
  recursive subroutine unstable_config_final (object)
    class(unstable_config_t), intent(inout) :: object
    integer :: i
    if (allocated (object%decay_config)) then
       do i = 1, size (object%decay_config)
          call object%decay_config(i)%final ()
       end do
    end if
  end subroutine unstable_config_final

  recursive subroutine unstable_config_write (object, unit, indent, verbose)
    class(unstable_config_t), intent(in) :: object
    integer, intent(in), optional :: unit, indent
    logical, intent(in), optional :: verbose
    integer :: u, i, ind
    logical :: verb
    u = output_unit (unit)
    ind = 0;  if (present (indent))  ind = indent
    verb = .true.;  if (present (verbose))  verb = verbose
    call write_indent (u, ind)
    write (u, "('+',1x,A,1x,A)")  "Unstable:", &
         char (flavor_get_name (object%flv))
    call write_indent (u, ind)
    write (u, 1)  "total width =", object%integral
    call write_indent (u, ind)
    write (u, 1)  "error (abs) =", object%abs_error
    call write_indent (u, ind)
    write (u, 1)  "error (rel) =", object%rel_error
1   format (4x,A,ES19.12)
    if (verb .and. allocated (object%decay_config)) then
       do i = 1, size (object%decay_config)
          call object%decay_config(i)%write (u, ind + 1)
       end do
    end if
  end subroutine unstable_config_write
  
  subroutine unstable_config_init (config, flv)
    class(unstable_config_t), intent(out) :: config
    type(flavor_t), intent(in) :: flv
    config%flv = flv
  end subroutine unstable_config_init
  
  subroutine unstable_config_init_decays (unstable, n_decays)
    class(unstable_config_t), intent(inout) :: unstable
    integer, intent(in) :: n_decays
    allocate (unstable%decay_config (n_decays))
  end subroutine unstable_config_init_decays
  
  recursive subroutine unstable_config_compute (unstable)
    class(unstable_config_t), intent(inout) :: unstable
    integer :: i
    do i = 1, size (unstable%decay_config)
       call unstable%decay_config(i)%compute ()
    end do
    unstable%integral = sum (unstable%decay_config%integral)
    if (unstable%integral <= 0) then
       call unstable%write ()
       call msg_fatal ("Decay configuration: computed total width is zero")
    end if
    unstable%abs_error = sqrt (sum (unstable%decay_config%abs_error ** 2))
    unstable%rel_error = unstable%abs_error / unstable%integral
    call unstable%selector%init (unstable%decay_config%integral)
    do i = 1, size (unstable%decay_config)
       unstable%decay_config(i)%weight &
            = unstable%selector%get_weight (i)
    end do
  end subroutine unstable_config_compute
    
  recursive subroutine unstable_final (object)
    class(unstable_t), intent(inout) :: object
    integer :: i
    if (allocated (object%decay)) then
       do i = 1, size (object%decay)
          call object%decay(i)%final ()
       end do
    end if
  end subroutine unstable_final
  
  recursive subroutine unstable_write (object, unit, indent)
    class(unstable_t), intent(in) :: object
    integer, intent(in), optional :: unit, indent
    integer :: u, i, ind
    u = output_unit (unit)
    ind = 0;  if (present (indent))  ind = indent
    call object%config%write (u, ind, verbose=.false.)
    call write_indent (u, ind)
    if (object%mode > 0) then
       write (u, "(3x,A,I0)") " Decay mode  = ", object%mode
       call object%decay(object%mode)%write (u, ind + 1)
    else
       write (u, "(3x,A)")  " Decay mode  = [undefined]"
    end if
  end subroutine unstable_write
    
  recursive subroutine unstable_write_process_instances (unstable, unit)
    class(unstable_t), intent(in) :: unstable
    integer, intent(in), optional :: unit
    if (unstable%mode > 0) then
       call unstable%decay(unstable%mode)%write_process_instances (unit)
    end if
  end subroutine unstable_write_process_instances
  
  recursive subroutine unstable_init (unstable, config)
    class(unstable_t), intent(out) :: unstable
    type(unstable_config_t), intent(in), target :: config
    integer :: i
    unstable%config => config
    allocate (unstable%decay (size (config%decay_config)))
    do i = 1, size (config%decay_config)
       call unstable%decay(i)%init (config%decay_config(i))
    end do
  end subroutine unstable_init
  
  subroutine write_indent (unit, indent)
    integer, intent(in) :: unit
    integer, intent(in), optional :: indent
    if (present (indent)) then
       write (unit, "(A)", advance="no")  repeat ("  ", indent)
    end if
  end subroutine write_indent


  subroutine decays_test (u, results)
    integer, intent(in) :: u
    type(test_results_t), intent(inout) :: results
    call test (decays_1, "decays_1", &
         "branching and decay configuration", &
         u, results)
    call test (decays_2, "decays_2", &
         "cascade decay configuration", &
         u, results)
    call test (decays_3, "decays_3", &
         "associate process", &
         u, results)
    call test (decays_4, "decays_4", &
         "decay instance", &
         u, results)
  end subroutine decays_test
  
  subroutine decays_1 (u)
    integer, intent(in) :: u
    type(os_data_t) :: os_data
    type(model_t), pointer :: model
    type(model_list_t) :: model_list
    type(flavor_t) :: flv_h
    type(flavor_t), dimension(2,1) :: flv_hbb, flv_hgg
    type(unstable_config_t), allocatable :: unstable

    write (u, "(A)")  "* Test output: decays_1"
    write (u, "(A)")  "*   Purpose: Set up branching and decay configuration"
    write (u, "(A)")

    write (u, "(A)")  "* Initialize environment"
    write (u, "(A)")

    call syntax_model_file_init ()
    call os_data_init (os_data)
    call model_list%read_model (var_str ("SM"), &
         var_str ("SM.mdl"), os_data, model)

    call flavor_init (flv_h, 25, model)
    call flavor_init (flv_hbb(:,1), [5, -5], model)
    call flavor_init (flv_hgg(:,1), [22, 22], model)

    write (u, "(A)")  "* Set up branching and decay"
    write (u, "(A)")

    allocate (unstable)
    unstable%flv = flv_h
    call unstable%init_decays (2)
    
    associate (decay => unstable%decay_config(1))
      call decay%init (var_str ("h_bb"), 1)
      call decay%init_term (1, flv_hbb, stable = [.true., .true.])
      decay%integral = 1.234e-3_default
      decay%abs_error = decay%integral * .02_default
    end associate
    
    associate (decay => unstable%decay_config(2))
      call decay%init (var_str ("h_gg"), 1)
      call decay%init_term (1, flv_hgg, stable = [.true., .true.])
      decay%integral = 3.085e-4_default
      decay%abs_error = decay%integral * .08_default
    end associate
    
    call unstable%compute ()
    call unstable%write (u)
    
    write (u, "(A)")
    write (u, "(A)")  "* Cleanup"

    call unstable%final ()
    call model_list%final ()
    call syntax_model_file_final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: decays_1"
    
  end subroutine decays_1
  
  subroutine decays_2 (u)
    integer, intent(in) :: u
    type(os_data_t) :: os_data
    type(model_t), pointer :: model
    type(model_list_t) :: model_list
    type(flavor_t) :: flv_h, flv_wp, flv_wm
    type(flavor_t), dimension(2,1) :: flv_hww, flv_wud, flv_wen
    type(unstable_config_t), allocatable :: unstable

    write (u, "(A)")  "* Test output: decays_2"
    write (u, "(A)")  "*   Purpose: Set up cascade branching"
    write (u, "(A)")

    write (u, "(A)")  "* Initialize environment"
    write (u, "(A)")

    call syntax_model_file_init ()
    call os_data_init (os_data)
    call model_list%read_model (var_str ("SM"), &
         var_str ("SM.mdl"), os_data, model)

    call flavor_init (flv_h, 25, model)
    call flavor_init (flv_hww(:,1), [24, -24], model)
    call flavor_init (flv_wp, 24, model)
    call flavor_init (flv_wm,-24, model)
    call flavor_init (flv_wud(:,1), [2, -1], model)
    call flavor_init (flv_wen(:,1), [-11, 12], model)
    

    write (u, "(A)")  "* Set up branching and decay"
    write (u, "(A)")

    allocate (unstable)
    unstable%flv = flv_h
    call unstable%init_decays (1)
    
    associate (decay => unstable%decay_config(1))

      call decay%init (var_str ("h_ww"), 1)
      decay%integral = 1.e-3_default
      decay%abs_error = decay%integral * .01_default

      call decay%init_term (1, flv_hww, stable = [.false., .true.])

      select type (w => decay%term_config(1)%prt(1)%c)
      type is (unstable_config_t)

         call w%init_decays (2)
         
         associate (w_decay => w%decay_config(1))
           call w_decay%init (var_str ("w_ud"), 1)
           w_decay%integral = 2._default
           call w_decay%init_term (1, flv_wud, stable = [.true., .true.])
         end associate
         associate (w_decay => w%decay_config(2))
           call w_decay%init (var_str ("w_en"), 1)
           w_decay%integral = 1._default
           call w_decay%init_term (1, flv_wen, stable = [.true., .true.])
         end associate
         call w%compute ()

      end select
      
    end associate
    
    call unstable%compute ()
    call unstable%write (u)
    
    write (u, "(A)")
    write (u, "(A)")  "* Cleanup"

    call unstable%final ()
    call model_list%final ()
    call syntax_model_file_final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: decays_2"
    
  end subroutine decays_2
  
  subroutine decays_3 (u)
    integer, intent(in) :: u
    type(os_data_t) :: os_data
    type(model_list_t) :: model_list
    type(model_t), pointer :: model
    type(process_library_t), target :: lib
    type(string_t) :: procname, run_id
    type(var_list_t), pointer :: var_list
    class(prc_core_t), allocatable :: core_template
    class(mci_t), allocatable :: mci_template
    class(phs_config_t), allocatable :: phs_config_template
    type(process_t), allocatable, target :: process
    type(process_instance_t), allocatable, target :: process_instance
    type(unstable_config_t), allocatable :: unstable

    write (u, "(A)")  "* Test output: decays_3"
    write (u, "(A)")  "*   Purpose: Connect a decay configuration &
         &with a process"
    write (u, "(A)")

    write (u, "(A)")  "* Initialize environment and integrate process"
    write (u, "(A)")

    call syntax_model_file_init ()
    call os_data_init (os_data)
    call syntax_model_file_init ()
    call model_list%read_model (var_str ("Test"), &
         var_str ("Test.mdl"), os_data, model)

    procname = "decays_3_p"
    run_id = "decays_3"

    call prc_test_create_library (procname, lib, scattering = .false., &
         decay = .true.)
    var_list => model_get_var_list_ptr (model)
    call var_list_set_real (var_list, &
         var_str ("ff"), 0.4_default, is_known = .true.)
    call model_parameters_update (model)

    call reset_interaction_counter ()
    allocate (process)
    call process%init (procname, run_id, lib, os_data, model_list)
    
    allocate (test_t :: core_template)
    allocate (mci_midpoint_t :: mci_template)
    allocate (phs_single_config_t :: phs_config_template)
    call process%init_component &
         (1, core_template, mci_template, phs_config_template)

    call process%setup_beams ()
    call process%configure_phs ()
    call process%setup_mci ()
    call process%setup_terms ()

    allocate (process_instance)
    call process_instance%init (process)
    call process%integrate (process_instance, 1, n_it=1, n_calls=100)
    call process%final_integration (1)
    call process_instance%final ()
    deallocate (process_instance)

    call process%write (u)
    
    write (u, "(A)")
    write (u, "(A)")  "* Set up branching and decay"
    write (u, "(A)")

    allocate (unstable)
    call flavor_init (unstable%flv, 25, model)
    call unstable%init_decays (1)
    
    write (u, "(A)")  "* Connect decay with process object"
    write (u, "(A)")

    associate (decay => unstable%decay_config(1))
      call decay%connect (process)
    end associate
    
    call unstable%compute ()
    call unstable%write (u)
    
    write (u, "(A)")
    write (u, "(A)")  "* Cleanup"

    call unstable%final ()
    call model_list%final ()
    call syntax_model_file_final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: decays_3"
    
  end subroutine decays_3

  subroutine decays_4 (u)
    integer, intent(in) :: u
    type(os_data_t) :: os_data
    type(model_list_t) :: model_list
    type(model_t), pointer :: model
    type(process_library_t), target :: lib
    type(string_t) :: procname, run_id
    type(var_list_t), pointer :: var_list
    class(prc_core_t), allocatable :: core_template
    class(mci_t), allocatable :: mci_template
    class(phs_config_t), allocatable :: phs_config_template
    type(process_t), allocatable, target :: process
    type(process_instance_t), allocatable, target :: process_instance
    type(unstable_config_t), allocatable, target :: unstable
    type(unstable_t), allocatable :: instance

    write (u, "(A)")  "* Test output: decays_4"
    write (u, "(A)")  "*   Purpose: Create a decay process and evaluate &
         &an instance"
    write (u, "(A)")

    write (u, "(A)")  "* Initialize environment, process, &
         &and decay configuration"
    write (u, "(A)")

    call syntax_model_file_init ()
    call os_data_init (os_data)
    call syntax_model_file_init ()
    call model_list%read_model (var_str ("Test"), &
         var_str ("Test.mdl"), os_data, model)

    procname = "decays_4_p"
    run_id = "decays_4"

    call prc_test_create_library (procname, lib, scattering = .false., &
         decay = .true.)
    var_list => model_get_var_list_ptr (model)
    call var_list_set_real (var_list, &
         var_str ("ff"), 0.4_default, is_known = .true.)
    call model_parameters_update (model)

    call reset_interaction_counter ()
    allocate (process)
    call process%init (procname, run_id, lib, os_data, model_list)
    
    allocate (test_t :: core_template)
    allocate (mci_midpoint_t :: mci_template)
    allocate (phs_single_config_t :: phs_config_template)
    call process%init_component &
         (1, core_template, mci_template, phs_config_template)

    call process%setup_beams ()
    call process%configure_phs ()
    call process%setup_mci ()
    call process%setup_terms ()

    allocate (process_instance)
    call process_instance%init (process)
    call process%integrate (process_instance, 1, n_it=1, n_calls=100)
    call process%final_integration (1)
    call process_instance%final ()
    deallocate (process_instance)

    allocate (unstable)
    call flavor_init (unstable%flv, 25, model)
    call unstable%init_decays (1)
    
    associate (decay => unstable%decay_config(1))
      call decay%connect (process)
    end associate
    
    call unstable%compute ()

    allocate (instance)
    call instance%init (unstable)
    
    instance%mode = 1
    instance%decay(1)%selected = 1
    
    call instance%write (u)
    write (u, *)
    call instance%write_process_instances (u)
    
    write (u, "(A)")
    write (u, "(A)")  "* Cleanup"

    call instance%final ()
    call unstable%final ()
    call model_list%final ()
    call syntax_model_file_final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: decays_4"
    
  end subroutine decays_4


end module decays

