! WHIZARD 2.2.0_alpha-4 Nov 15 2013
! 
! Copyright (C) 1999-2013 by 
!     Wolfgang Kilian <kilian@physik.uni-siegen.de>
!     Thorsten Ohl <ohl@physik.uni-wuerzburg.de>
!     Juergen Reuter <juergen.reuter@desy.de>
!     
!     with contributions by 
!     Christian Speckner <cnspeckn@googlemail.com> 
!     and  Felix Braam, Sebastian Schmidt, Daniel Wiesler 
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module dispatch
  
  use kinds, only: default !NODEP!
  use iso_varying_string, string_t => varying_string !NODEP!
  use limits, only: MZ_REF, ALPHA_QCD_MZ_REF !NODEP!
  use constants, only: PI !NODEP!
  use file_utils !NODEP!
  use diagnostics !NODEP!
  use os_interface
  use unit_tests
  use sm_qcd
  use variables
  use flavors
  use interactions
  use models
  use process_constants
  use pdf_builtin !NODEP!
  use sf_mappings
  use sf_base
  use sf_pdf_builtin
  use sf_lhapdf
  use sf_circe1
  use sf_circe2
  use sf_lumi_linker
  use sf_isr
  use sf_epa
  use sf_ewa
  use sf_escan
  use sf_beam_events
  use sf_user
  use rng_base
  use rng_tao
  use mci_base
  use mci_midpoint
  use mci_vamp
  use phs_base
  use mappings
  use phs_forests
  use phs_single
  use phs_wood
  use prc_core_def
  use prc_test
  use beams
  use prc_omega
  use prc_core
  use processes
  use beam_structures
  use eio_base
  use eio_raw
  use eio_lhef
  use eio_hepmc
  use eio_stdhep
  use eio_ascii
  use eio_weights
  use rt_data
  
  implicit none
  private

  public :: dispatch_core_def
  public :: dispatch_core
  public :: dispatch_core_update
  public :: dispatch_core_restore
  public :: dispatch_mci
  public :: dispatch_phs
  public :: dispatch_rng
  public :: dispatch_sf_data
  public :: dispatch_sf_config
  public :: dispatch_eio
  public :: dispatch_qcd
  public :: dispatch_test

contains
  
  subroutine dispatch_core_def (core_def, prt_in, prt_out, global)
    class(prc_core_def_t), allocatable, intent(inout) :: core_def
    type(string_t), dimension(:), intent(in) :: prt_in
    type(string_t), dimension(:), intent(in) :: prt_out
    type(rt_data_t), intent(in) :: global
    type(string_t) :: method
    type(string_t) :: model_name
    type(string_t) :: restrictions
    logical :: openmp_support
    logical :: report_progress
    type(string_t) :: extra_options
    associate (var_list => global%var_list)
      method = var_list_get_sval (var_list, var_str ("$method"))
      model_name = var_list_get_sval (var_list, var_str ("$model_name"))
      select case (char (method))
      case ("unit_test")
         allocate (prc_test_def_t :: core_def)
         select type (core_def)
         type is (prc_test_def_t)
            call core_def%init (model_name, prt_in, prt_out)
         end select
      case ("omega")
         restrictions = var_list_get_sval (var_list, &
              var_str ("$restrictions"))
         openmp_support = var_list_get_lval (var_list, &
              var_str ("?omega_openmp"))
         report_progress = var_list_get_lval (var_list, &
              var_str ("?report_progress"))
         extra_options = var_list_get_sval (var_list, &
              var_str ("$omega_flags"))
         allocate (omega_def_t :: core_def)
         select type (core_def)
         type is (omega_def_t)
            call core_def%init (model_name, prt_in, prt_out, &
                 restrictions, openmp_support, report_progress, extra_options)
         end select
      case default
         call msg_fatal ("Process configuration: method '" &
              // char (method) // "' not implemented")
      end select
    end associate
  end subroutine dispatch_core_def
    
  subroutine dispatch_core (core, core_def, model, helicity_selection, qcd)
    class(prc_core_t), allocatable, intent(inout) :: core
    class(prc_core_def_t), intent(in) :: core_def
    type(model_t), intent(in), target, optional :: model
    type(helicity_selection_t), intent(in), optional :: helicity_selection
    type(qcd_t), intent(in), optional :: qcd
    select type (core_def)
    type is (prc_test_def_t)
       allocate (test_t :: core)
    type is (omega_def_t)
       allocate (prc_omega_t :: core)
       select type (core)
       type is (prc_omega_t)
          call core%set_parameters (model, helicity_selection, qcd)
       end select
    class default
       call msg_bug ("Process core: unexpected process definition type")
    end select
  end subroutine dispatch_core

  subroutine dispatch_core_update (core, model, helicity_selection, qcd, &
       saved_core)
    class(prc_core_t), allocatable, intent(inout) :: core
    type(model_t), intent(in), optional, target :: model
    type(helicity_selection_t), intent(in), optional :: helicity_selection
    type(qcd_t), intent(in), optional :: qcd
    class(prc_core_t), allocatable, intent(inout), optional :: saved_core
    if (present (saved_core)) then
       allocate (saved_core, source = core)
    end if
    select type (core)
    type is (test_t)
    type is (prc_omega_t)
       call core%set_parameters (model, helicity_selection, qcd)
       call core%activate_parameters ()
    class default
       call msg_bug ("Process core update: unexpected process definition type")
    end select
  end subroutine dispatch_core_update

  subroutine dispatch_core_restore (core, saved_core)
    class(prc_core_t), allocatable, intent(inout) :: core
    class(prc_core_t), allocatable, intent(inout) :: saved_core
    call move_alloc (from = saved_core, to = core)
    select type (core)
    type is (test_t)
    type is (prc_omega_t)
       call core%activate_parameters ()
    class default
       call msg_bug ("Process core restore: unexpected process definition type")
    end select
  end subroutine dispatch_core_restore

  subroutine dispatch_mci (mci, global, process_id)
    class(mci_t), allocatable, intent(inout) :: mci
    type(rt_data_t), intent(in) :: global
    type(string_t), intent(in) :: process_id
    type(string_t) :: run_id
    type(string_t) :: integration_method
    type(grid_parameters_t) :: grid_par
    logical :: rebuild_grids
    integration_method = var_list_get_sval (global%var_list, &
         var_str ("$integration_method"))
    select case (char (integration_method))
    case ("midpoint")
       allocate (mci_midpoint_t :: mci)
    case ("vamp", "default")
       associate (var_list => global%var_list)
         grid_par%threshold_calls = &
              var_list_get_ival (var_list, var_str ("threshold_calls"))
         grid_par%min_calls_per_channel = &
              var_list_get_ival (var_list, var_str ("min_calls_per_channel"))
         grid_par%min_calls_per_bin = &
              var_list_get_ival (var_list, var_str ("min_calls_per_bin"))
         grid_par%min_bins = &
              var_list_get_ival (var_list, var_str ("min_bins"))
         grid_par%max_bins = &
              var_list_get_ival (var_list, var_str ("max_bins"))
         grid_par%stratified = &
              var_list_get_lval (var_list, var_str ("?stratified"))
         grid_par%use_vamp_equivalences = &
              var_list_get_lval (var_list, var_str ("?use_vamp_equivalences"))
         grid_par%channel_weights_power = &
              var_list_get_rval (var_list, var_str ("channel_weights_power"))
         run_id = &
              var_list_get_sval (var_list, var_str ("$run_id"))
         rebuild_grids = &
              var_list_get_lval (var_list, var_str ("?rebuild_grids"))
       end associate
       allocate (mci_vamp_t :: mci)
       select type (mci)
       type is (mci_vamp_t)
          call mci%set_grid_parameters (grid_par)
          if (run_id /= "") then
             call mci%set_grid_filename (process_id, run_id)
          else
             call mci%set_grid_filename (process_id)
          end if
          call mci%set_rebuild_flag (rebuild_grids)
       end select
    case default
       call msg_fatal ("Integrator '" &
            // char (integration_method) // "' not implemented")
    end select
  end subroutine dispatch_mci
  
  subroutine dispatch_phs (phs, global, mapping_defaults, phs_par)
    class(phs_config_t), allocatable, intent(inout) :: phs
    type(rt_data_t), intent(in) :: global
    type(mapping_defaults_t), intent(in), optional :: mapping_defaults
    type(phs_parameters_t), intent(in), optional :: phs_par
    type(string_t) :: phs_method, phs_file
    logical :: use_equivalences
    integer :: u_phs
    logical :: exist
    phs_method = var_list_get_sval (global%var_list, &
         var_str ("$phs_method"))
    phs_file = var_list_get_sval (global%var_list, &
         var_str ("$phs_file"))
    use_equivalences = var_list_get_lval (global%var_list, &
         var_str ("?use_vamp_equivalences"))
    select case (char (phs_method))
    case ("single")
       allocate (phs_single_config_t :: phs)
    case ("wood", "default")
       allocate (phs_wood_config_t :: phs)
       select type (phs)
       type is (phs_wood_config_t)
          if (phs_file /= "") then
             inquire (file = char (phs_file), exist = exist)
             if (exist) then
                call msg_message ("Phase space: reading configuration from '" &
                     // char (phs_file) // "'")
                u_phs = free_unit ()
                open (u_phs, file = char (phs_file), &
                     action = "read", status = "old")
                call phs%set_input (u_phs)
             else
                call msg_fatal ("Phase space: configuration file '" &
                     // char (phs_file) // "' not found")
             end if
          end if
          if (present (phs_par)) &
               call phs%set_parameters (phs_par)
          if (use_equivalences) &
               call phs%enable_equivalences ()
          if (present (mapping_defaults)) &
               call phs%set_mapping_defaults (mapping_defaults)
       end select
    case default
       call msg_fatal ("Phase space: parameterization method '" &
            // char (phs_method) // "' not implemented")
    end select
  end subroutine dispatch_phs
  
  subroutine dispatch_rng (rng, global)
    class(rng_t), allocatable, intent(inout) :: rng
    type(rt_data_t), intent(in) :: global
    type(string_t) :: rng_method
    rng_method = var_list_get_sval (global%var_list, &
         var_str ("$rng_method"))
    select case (char (rng_method))
    case ("unit_test")
       allocate (rng_test_t :: rng)
    case ("tao")
       allocate (rng_tao_t :: rng)
    case default
       call msg_fatal ("Random-number generator '" &
            // char (rng_method) // "' not implemented")
    end select
  end subroutine dispatch_rng
  
  subroutine dispatch_sf_data (data, sf_method, i_beam, global)
    class(sf_data_t), allocatable, intent(inout) :: data
    type(string_t), intent(in) :: sf_method
    integer, dimension(:), intent(in) :: i_beam
    type(rt_data_t), intent(inout) :: global
    type(model_t), pointer :: model
    type(string_t), dimension(2) :: prt_beam
    type(flavor_t), dimension(2) :: flv_beam
    real(default) :: sqrts, isr_alpha, isr_q_max, isr_mass
    integer :: isr_order
    logical :: isr_recoil
    real(default) :: epa_alpha, epa_x_min, epa_q_min, epa_e_max, epa_mass
    logical :: epa_recoil
    real(default) :: ewa_x_min, ewa_q_min, ewa_pt_max, ewa_mass
    logical :: ewa_keep_momentum, ewa_keep_energy   
    type(string_t) :: pdf_name, pdf_path
    type(string_t) :: lhapdf_dir, lhapdf_file
    integer :: lhapdf_member, lhapdf_photon_scheme
    model => global%model
    sqrts = var_list_get_rval (global%var_list, var_str ("sqrts"))
    prt_beam = global%beam_structure%get_prt ()
    call flavor_init (flv_beam, prt_beam, model)
    associate (var_list => global%var_list)
      select case (char (sf_method))
      case ("sf_test_0", "sf_test_1")
         allocate (sf_test_data_t :: data)
         select type (data)
         type is (sf_test_data_t)
            select case (char (sf_method))
            case ("sf_test_0");  call data%init (model, flv_beam(i_beam(1)))
            case ("sf_test_1");  call data%init (model, flv_beam(i_beam(1)), &
                 mode = 1)
            end select
         end select
      case ("pdf_builtin")
         allocate (pdf_builtin_data_t :: data)
         select type (data)
         type is (pdf_builtin_data_t)
            pdf_name = &
                 var_list_get_sval (var_list, var_str ("$pdf_builtin_set"))
            call data%init (global%pdf_builtin_status, &
                 model, flv_beam(i_beam(1)), &
                 name = pdf_name, &
                 path = global%os_data%pdf_builtin_datapath)
         end select
      case ("lhapdf")
         allocate (lhapdf_data_t :: data)
         lhapdf_dir = &
              var_list_get_sval (var_list, var_str ("$lhapdf_dir"))  
         lhapdf_file = &
              var_list_get_sval (var_list, var_str ("$lhapdf_file")) 
         lhapdf_member = &
              var_list_get_ival (var_list, var_str ("lhapdf_member"))
         lhapdf_photon_scheme = &
              var_list_get_ival (var_list, var_str ("lhapdf_photon_scheme"))             
         select type (data)
         type is (lhapdf_data_t)
            call data%init &
                 (global%lhapdf_status, model, flv_beam(i_beam(1)), &
                  lhapdf_dir, lhapdf_file, lhapdf_member, &
                  lhapdf_photon_scheme)
         end select
      case ("isr")
         allocate (isr_data_t :: data)
         isr_alpha = var_list_get_rval (var_list, var_str ("isr_alpha"))
         if (isr_alpha == 0) then
            isr_alpha = (var_list_get_rval (var_list, var_str ("ee"))) &
                 ** 2 / (4 * PI)
         end if
         isr_q_max = var_list_get_rval (var_list, var_str ("isr_q_max"))
         if (isr_q_max == 0) then
            isr_q_max = sqrts
         end if
         isr_mass   = var_list_get_rval (var_list, var_str ("isr_mass"))
         isr_order  = var_list_get_ival (var_list, var_str ("isr_order"))
         isr_recoil = var_list_get_lval (var_list, var_str ("?isr_recoil")) 
         select type (data)
         type is (isr_data_t)
            call data%init &
                 (model, flv_beam (i_beam(1)), isr_alpha, isr_q_max, &
                 isr_mass, isr_order, isr_recoil)
            call data%check ()
         end select
      case ("epa")
         allocate (epa_data_t :: data)
         epa_alpha = var_list_get_rval (var_list, var_str ("epa_alpha"))
         if (epa_alpha == 0) then
            epa_alpha = (var_list_get_rval (var_list, var_str ("ee"))) &
                 ** 2 / (4 * PI)
         end if         
         epa_x_min = var_list_get_rval (var_list, var_str ("epa_x_min"))
         epa_q_min = var_list_get_rval (var_list, var_str ("epa_q_min"))
         epa_e_max = var_list_get_rval (var_list, var_str ("epa_e_max"))
         if (epa_e_max == 0) then
            epa_e_max = sqrts
         end if
         epa_mass   = var_list_get_rval (var_list, var_str ("epa_mass"))
         epa_recoil = var_list_get_lval (var_list, var_str ("?epa_recoil"))
         select type (data)            
         type is (epa_data_t)
            call data%init &
                 (model, flv_beam (i_beam(1)), epa_alpha, epa_x_min, &
                 epa_q_min, epa_e_max, epa_mass, epa_recoil)
            call data%check ()
         end select
      case ("ewa")
         allocate (ewa_data_t :: data)
         ewa_x_min = var_list_get_rval (var_list, var_str ("ewa_x_min"))
         ewa_q_min = var_list_get_rval (var_list, var_str ("ewa_q_min"))
         ewa_pt_max = var_list_get_rval (var_list, var_str ("ewa_pt_max")) 
         if (ewa_pt_max == 0) then
            ewa_pt_max = sqrts
         end if
         ewa_mass = var_list_get_rval (var_list, var_str ("ewa_mass"))  
         ewa_keep_momentum = var_list_get_lval (var_list, &
              var_str ("?ewa_keep_momentum"))
         ewa_keep_energy = var_list_get_lval (var_list, &
                    var_str ("?ewa_keep_energy"))                  
         if (ewa_keep_momentum .and. ewa_keep_energy) &
              call msg_fatal (" EWA cannot conserve both energy " &
                 // "and momentum.")          
         select type (data)
         type is (ewa_data_t)
            if (ewa_mass /= 0) then
               call data%init &
                    (model, flv_beam (i_beam(1)), ewa_x_min, ewa_q_min, &
                    ewa_pt_max, sqrts, ewa_keep_momentum, &
                    ewa_keep_energy, ewa_mass)
            else
               call data%init &
                    (model, flv_beam (i_beam(1)), ewa_x_min, ewa_q_min, &
                    ewa_pt_max, sqrts, ewa_keep_momentum, &
                    ewa_keep_energy)
            end if
            call data%check ()
         end select
      case ("energy_scan")
      case default
         call msg_bug ("Structure function '" &
              // char (sf_method) // "' not implemented yet")
      end select
    end associate
  end subroutine dispatch_sf_data
  
  function strfun_mode (name) result (n)
    type(string_t), intent(in) :: name
    integer :: n
    select case (char (name))
    case ("none")
       n = 0
    case ("sf_test_0", "sf_test_1")
       n = 1
    case ("pdf_builtin","lhapdf","isr","epa","ewa")
       n = 1
    case default
       call msg_bug ("Structure function '" // char (name) &
            // "' not supported yet")
    end select
  end function strfun_mode
    
  subroutine dispatch_sf_config (sf_config, sf_channel, sf_string, global)
    type(sf_config_t), dimension(:), allocatable, intent(out) :: sf_config
    type(sf_channel_t), dimension(:), allocatable, intent(out) :: sf_channel
    type(string_t), intent(out) :: sf_string
    type(rt_data_t), intent(inout) :: global
    type(beam_structure_t) :: beam_structure
    class(sf_data_t), allocatable :: sf_data
    integer :: n_record, i
    beam_structure = global%beam_structure
    call beam_structure%expand (strfun_mode)
    n_record = beam_structure%get_n_record ()
    allocate (sf_config (n_record))
    do i = 1, n_record
       call dispatch_sf_data (sf_data, &
            beam_structure%get_name (i), &
            beam_structure%get_i_entry (i), &
            global)
       call sf_config(i)%init (beam_structure%get_i_entry (i), sf_data)
       deallocate (sf_data)
    end do
    sf_string = beam_structure%to_string (sf_only = .true.)
    select case (char (sf_string))
    !!! JRR: WK please check
    !!! What about this? How for the other SFs?
    !!! Single structure function not covered yet
    case ("pdf_builtin, none => none, pdf_builtin")
       allocate (sf_channel (1))
       call sf_channel(1)%init (2)
       call sf_channel(1)%set_s_mapping ([1, 2], power = 2._default)
    case ("lhapdf, none => none, lhapdf")
       allocate (sf_channel (1))
       call sf_channel(1)%init (2)
       call sf_channel(1)%set_s_mapping ([1, 2], power = 2._default)
    case ("isr, none => none, isr")
       allocate (sf_channel (1))
       call sf_channel(1)%init(2)
    case ("epa, none => none, epa")
       allocate (sf_channel (1))
       call sf_channel(1)%init(2)       
    end select
  end subroutine dispatch_sf_config
    
  subroutine dispatch_eio (eio, method, global)
    class(eio_t), intent(inout), allocatable :: eio
    type(string_t), intent(in) :: method
    type(rt_data_t), intent(in) :: global
    logical :: check, keep_beams, recover_beams
    logical :: write_sqme_prc, write_sqme_ref, write_sqme_alt
    type(string_t) :: lhef_version, lhef_extension
    type(string_t) :: extension_default, extension_debug, extension_hepmc, &
         extension_lha, extension_hepevt, extension_ascii_short, &
         extension_ascii_long, extension_athena, extension_mokka, &
         extension_stdhep, extension_stdhep_up, extension_raw, &
         extension_hepevt_verb, extension_lha_verb
    select case (char (method))
    case ("raw")
       allocate (eio_raw_t :: eio)
       select type (eio)
       type is (eio_raw_t)
          check = var_list_get_lval &
               (global%var_list, var_str ("?check_event_file"))
          extension_raw = &
               var_list_get_sval (global%var_list, var_str ("$extension_raw"))
          call eio%set_parameters (check, extension_raw)
       end select
    case ("lhef")
       allocate (eio_lhef_t :: eio)
       select type (eio)
       type is (eio_lhef_t)
          keep_beams = &
               var_list_get_lval (global%var_list, var_str ("?keep_beams"))
          recover_beams = &
               var_list_get_lval (global%var_list, var_str ("?recover_beams"))
          lhef_version = &
               var_list_get_sval (global%var_list, var_str ("$lhef_version"))
          lhef_extension = &
               var_list_get_sval (global%var_list, var_str ("$lhef_extension"))
          write_sqme_prc = var_list_get_lval (global%var_list, &
               var_str ("?lhef_write_sqme_prc"))
          write_sqme_ref = var_list_get_lval (global%var_list, &
               var_str ("?lhef_write_sqme_ref"))
          write_sqme_alt = var_list_get_lval (global%var_list, &
               var_str ("?lhef_write_sqme_alt"))
          call eio%set_parameters (keep_beams, recover_beams, &
               char (lhef_version), lhef_extension, &
               write_sqme_ref, write_sqme_prc, write_sqme_alt)
       end select
    case ("hepmc")
       allocate (eio_hepmc_t :: eio)
       select type (eio)
       type is (eio_hepmc_t)
          keep_beams = &
               var_list_get_lval (global%var_list, var_str ("?keep_beams"))
          extension_hepmc = var_list_get_sval ( &
               global%var_list, var_str ("$extension_hepmc"))          
          call eio%set_parameters (keep_beams, extension_hepmc)
       end select
    case ("stdhep")
       allocate (eio_stdhep_hepevt_t :: eio)
       select type (eio)
       type is (eio_stdhep_hepevt_t)                   
          keep_beams = &
               var_list_get_lval (global%var_list, var_str ("?keep_beams"))
          extension_stdhep = var_list_get_sval ( &
               global%var_list, var_str ("$extension_stdhep"))          
          call eio%set_parameters (keep_beams, extension_stdhep)          
       end select
    case ("stdhep_up")
       allocate (eio_stdhep_hepeup_t :: eio)
       select type (eio)
       type is (eio_stdhep_hepeup_t)          
          keep_beams = &
               var_list_get_lval (global%var_list, var_str ("?keep_beams"))
          extension_stdhep_up = var_list_get_sval ( &
               global%var_list, var_str ("$extension_stdhep_up"))          
          call eio%set_parameters (keep_beams, extension_stdhep_up)          
       end select       
    case ("ascii")   
       allocate (eio_ascii_ascii_t :: eio)
       select type (eio)
       type is (eio_ascii_ascii_t)
          keep_beams = &
               var_list_get_lval (global%var_list, var_str ("?keep_beams"))
          extension_default = var_list_get_sval ( &
               global%var_list, var_str ("$extension_default"))          
          call eio%set_parameters (keep_beams, extension_default)
       end select       
    case ("athena")   
       allocate (eio_ascii_athena_t :: eio)
       select type (eio)
       type is (eio_ascii_athena_t)
          keep_beams = &
               var_list_get_lval (global%var_list, var_str ("?keep_beams"))
          extension_athena = var_list_get_sval ( &
               global%var_list, var_str ("$extension_athena"))
          call eio%set_parameters (keep_beams, extension_athena)
       end select              
    case ("debug")   
       allocate (eio_ascii_debug_t :: eio)
       select type (eio)
       type is (eio_ascii_debug_t)
          keep_beams = &
               var_list_get_lval (global%var_list, var_str ("?keep_beams"))
          extension_debug = var_list_get_sval ( &
               global%var_list, var_str ("$extension_debug"))          
          call eio%set_parameters (keep_beams, extension_debug)
       end select
    case ("hepevt")   
       allocate (eio_ascii_hepevt_t :: eio)
       select type (eio)
       type is (eio_ascii_hepevt_t)
          keep_beams = &
               var_list_get_lval (global%var_list, var_str ("?keep_beams"))
          extension_hepevt = var_list_get_sval ( &
               global%var_list, var_str ("$extension_hepevt"))          
          call eio%set_parameters (keep_beams, extension_hepevt)
       end select              
    case ("hepevt_verb")   
       allocate (eio_ascii_hepevt_verb_t :: eio)
       select type (eio)
       type is (eio_ascii_hepevt_verb_t)
          keep_beams = &
               var_list_get_lval (global%var_list, var_str ("?keep_beams"))
          extension_hepevt_verb = var_list_get_sval ( &
               global%var_list, var_str ("$extension_hepevt_verb"))          
          call eio%set_parameters (keep_beams, extension_hepevt_verb)
       end select                     
    case ("lha")   
       allocate (eio_ascii_lha_t :: eio)
       select type (eio)
       type is (eio_ascii_lha_t)
          keep_beams = &
               var_list_get_lval (global%var_list, var_str ("?keep_beams"))
          extension_lha = var_list_get_sval ( &
               global%var_list, var_str ("$extension_lha"))          
          call eio%set_parameters (keep_beams, extension_lha)
       end select                     
    case ("lha_verb")   
       allocate (eio_ascii_lha_verb_t :: eio)
       select type (eio)
       type is (eio_ascii_lha_verb_t)
          keep_beams = &
               var_list_get_lval (global%var_list, var_str ("?keep_beams"))
          extension_lha_verb = var_list_get_sval ( &
               global%var_list, var_str ("$extension_lha_verb"))          
          call eio%set_parameters (keep_beams, extension_lha_verb)
       end select                            
    case ("long")   
       allocate (eio_ascii_long_t :: eio)
       select type (eio)
       type is (eio_ascii_long_t)
          keep_beams = &
               var_list_get_lval (global%var_list, var_str ("?keep_beams"))
          extension_ascii_long = var_list_get_sval ( &
               global%var_list, var_str ("$extension_ascii_long"))          
          call eio%set_parameters (keep_beams, extension_ascii_long)
       end select              
    case ("mokka")   
       allocate (eio_ascii_mokka_t :: eio)
       select type (eio)
       type is (eio_ascii_mokka_t)
          keep_beams = &
               var_list_get_lval (global%var_list, var_str ("?keep_beams"))
          extension_mokka = var_list_get_sval ( &
               global%var_list, var_str ("$extension_mokka"))          
          call eio%set_parameters (keep_beams, extension_mokka)
       end select                     
    case ("short")   
       allocate (eio_ascii_short_t :: eio)
       select type (eio)
       type is (eio_ascii_short_t)
          keep_beams = &
               var_list_get_lval (global%var_list, var_str ("?keep_beams"))
          extension_ascii_short = var_list_get_sval ( &
               global%var_list, var_str ("$extension_ascii_short"))          
          call eio%set_parameters (keep_beams, extension_ascii_short)
       end select                     
    case ("weight_stream")
       allocate (eio_weights_t :: eio)
    case default
       call msg_fatal ("Event I/O method '" // char (method) &
            // "' not implemented")
    end select
  end subroutine dispatch_eio
  
  subroutine dispatch_qcd (qcd, global)
    type(qcd_t), intent(inout) :: qcd
    type(rt_data_t), intent(inout), target :: global
    type(var_list_t), pointer :: var_list
    logical :: fixed, from_mz, from_pdf_builtin, from_lhapdf
    real(default) :: mz, alpha_val, lambda
    integer :: nf, order, lhapdf_member
    type(string_t) :: pdfset, lhapdf_dir, lhapdf_file
    var_list => global%var_list
    fixed = &
         var_list_get_lval (var_list, var_str ("?alpha_s_is_fixed"))
    from_mz = &
         var_list_get_lval (var_list, var_str ("?alpha_s_from_mz"))
    from_pdf_builtin = &
         var_list_get_lval (var_list, var_str ("?alpha_s_from_pdf_builtin"))
    from_lhapdf = &
         var_list_get_lval (var_list, var_str ("?alpha_s_from_lhapdf"))
    pdfset = &
         var_list_get_sval (var_list, var_str ("$pdf_builtin_set"))    
    lambda = &
         var_list_get_rval (var_list, var_str ("lambda_qcd"))
    nf = &
         var_list_get_ival (var_list, var_str ("alpha_s_nf"))
    order = &
         var_list_get_ival (var_list, var_str ("alpha_s_order"))
    lhapdf_dir = &
         var_list_get_sval (var_list, var_str ("$lhapdf_dir"))  
    lhapdf_file = &
         var_list_get_sval (var_list, var_str ("$lhapdf_file"))     
    lhapdf_member = &
         var_list_get_ival (var_list, var_str ("lhapdf_member"))         
    var_list => model_get_var_list_ptr (global%model)
    if (var_list_exists (var_list, var_str ("mZ"))) then
       mz = var_list_get_rval (var_list, var_str ("mZ"))
    else
       mz = MZ_REF
    end if
    if (var_list_exists (var_list, var_str ("alphas"))) then
       alpha_val = var_list_get_rval (var_list, var_str ("alphas"))
    else
       alpha_val = ALPHA_QCD_MZ_REF
    end if
    if (allocated (qcd%alpha))  deallocate (qcd%alpha)
    if (from_lhapdf .and. from_pdf_builtin) then
        call msg_fatal (" Mixing alphas evolution",  &
             (/ var_str (" from LHAPDF and builtin PDF is not permitted") /))
    end if 
    if (fixed) then
!!! JRR: WK please check: why commented out??       
!       allocate (alpha_qcd_fixed_t :: qcd%alpha)
    else if (from_mz) then
       allocate (alpha_qcd_from_scale_t :: qcd%alpha)
    else if (from_pdf_builtin) then
       allocate (alpha_qcd_pdf_builtin_t :: qcd%alpha)
    else if (from_lhapdf) then
       allocate (alpha_qcd_lhapdf_t :: qcd%alpha)
    else
       allocate (alpha_qcd_from_lambda_t :: qcd%alpha)
    end if
    if (allocated (qcd%alpha)) then
       select type (alpha => qcd%alpha)
       type is (alpha_qcd_fixed_t)
          alpha%val = alpha_val
       type is (alpha_qcd_from_scale_t)
          alpha%mu_ref = mz
          alpha%ref = alpha_val
          alpha%order = order
          alpha%nf = nf
       type is (alpha_qcd_from_lambda_t)
          alpha%lambda = lambda
          alpha%order = order
          alpha%nf = nf
       type is (alpha_qcd_pdf_builtin_t)
          call alpha%init (global%pdf_builtin_status, pdfset, &
               global%os_data%pdf_builtin_datapath)
       type is (alpha_qcd_lhapdf_t)
          call alpha%init (global%lhapdf_status, lhapdf_file, &
               lhapdf_member, lhapdf_dir)
       end select
       !!! JRR: WK please check
       !!! why is that commented out
!    else
!       call msg_bug ("dispatch_qcd: could not allocate QCD data")
    end if
  end subroutine dispatch_qcd
  

  subroutine dispatch_test (u, results)
    integer, intent(in) :: u
    type(test_results_t), intent(inout) :: results
    call test (dispatch_1, "dispatch_1", &
         "process configuration method", &
         u, results)
    call test (dispatch_2, "dispatch_2", &
         "process core", &
         u, results)
    call test (dispatch_3, "dispatch_3", &
         "integration method", &
         u, results)
    call test (dispatch_4, "dispatch_4", &
         "phase-space configuration", &
         u, results)
    call test (dispatch_5, "dispatch_5", &
         "random-number generator", &
         u, results)
    call test (dispatch_6, "dispatch_6", &
         "configure phase space using file", &
         u, results)
    call test (dispatch_7, "dispatch_7", &
         "structure-function data", &
         u, results)
    call test (dispatch_8, "dispatch_8", &
         "beam structure", &
         u, results)
    call test (dispatch_9, "dispatch_9", &
         "event I/O", &
         u, results)
    call test (dispatch_10, "dispatch_10", &
         "process core update", &
         u, results)
    call test (dispatch_11, "dispatch_11", &
         "QCD coupling", &
         u, results)
end subroutine dispatch_test

  subroutine dispatch_1 (u)
    integer, intent(in) :: u
    type(string_t), dimension(2) :: prt_in, prt_out
    type(rt_data_t), target :: global
    class(prc_core_def_t), allocatable :: core_def
    
    write (u, "(A)")  "* Test output: dispatch_1"
    write (u, "(A)")  "*   Purpose: select process configuration method"
    write (u, "(A)")

    call global%global_init ()
    call var_list_set_log (global%var_list, var_str ("?omega_openmp"), &
         .false., is_known = .true.)

    prt_in = [var_str ("a"), var_str ("b")]
    prt_out = [var_str ("c"), var_str ("d")]

    write (u, "(A)")  "* Allocate core_def as prc_test_def"

    call var_list_set_string (global%var_list, var_str ("$method"), &
         var_str ("unit_test"), is_known = .true.)
    call dispatch_core_def (core_def, prt_in, prt_out, global)
    select type (core_def)
    type is (prc_test_def_t)
       call core_def%write (u)
    end select
    
    deallocate (core_def)

    write (u, "(A)")
    write (u, "(A)")  "* Allocate core_def as omega_def"
    write (u, "(A)")

    call var_list_set_string (global%var_list, var_str ("$method"), &
         var_str ("omega"), is_known = .true.)
    call dispatch_core_def (core_def, prt_in, prt_out, global)
    select type (core_def)
    type is (omega_def_t)
       call core_def%write (u)
    end select
    
    call global%final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: dispatch_1"
    
  end subroutine dispatch_1
  
  subroutine dispatch_2 (u)
    integer, intent(in) :: u
    type(string_t), dimension(2) :: prt_in, prt_out
    type(rt_data_t), target :: global
    class(prc_core_def_t), allocatable :: core_def
    class(prc_core_t), allocatable :: core
    type(model_t), pointer :: model
    
    write (u, "(A)")  "* Test output: dispatch_2"
    write (u, "(A)")  "*   Purpose: select process configuration method"
    write (u, "(A)")  "             and allocate process core"
    write (u, "(A)")

    call syntax_model_file_init ()
    call global%global_init ()

    prt_in = [var_str ("a"), var_str ("b")]
    prt_out = [var_str ("c"), var_str ("d")]

    write (u, "(A)")  "* Allocate core as test_t"
    write (u, "(A)")

    call var_list_set_string (global%var_list, var_str ("$method"), &
         var_str ("unit_test"), is_known = .true.)
    call dispatch_core_def (core_def, prt_in, prt_out, global)
    call dispatch_core (core, core_def)
    select type (core)
    type is (test_t)
       call core%write (u)
    end select
    
    deallocate (core)
    deallocate (core_def)

    write (u, "(A)")
    write (u, "(A)")  "* Allocate core as prc_omega_t"
    write (u, "(A)")

    call var_list_set_string (global%var_list, var_str ("$method"), &
         var_str ("omega"), is_known = .true.)
    call dispatch_core_def (core_def, prt_in, prt_out, global)

    call global%read_model (var_str ("Test"), var_str ("Test.mdl"))

    call var_list_set_log (global%var_list, &
         var_str ("?helicity_selection_active"), &
         .true., is_known = .true.)
    call var_list_set_real (global%var_list, &
         var_str ("helicity_selection_threshold"), &
         1e9_default, is_known = .true.)
    call var_list_set_int (global%var_list, &
         var_str ("helicity_selection_cutoff"), &
         10, is_known = .true.)
    
    call dispatch_core (core, core_def, &
         global%model, global%get_helicity_selection ())
    call core_def%allocate_driver (core%driver, var_str (""))

    select type (core)
    type is (prc_omega_t)
       call core%write (u)
    end select
    
    call global%final ()
    call syntax_model_file_final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: dispatch_2"
    
  end subroutine dispatch_2
  
  subroutine dispatch_3 (u)
    integer, intent(in) :: u
    type(rt_data_t), target :: global
    class(mci_t), allocatable :: mci
    type(string_t) :: process_id
    
    write (u, "(A)")  "* Test output: dispatch_3"
    write (u, "(A)")  "*   Purpose: select integration method"
    write (u, "(A)")

    call global%global_init ()
    process_id = "dispatch_3"

    write (u, "(A)")  "* Allocate MCI as midpoint_t"
    write (u, "(A)")

    call var_list_set_string (global%var_list, &
         var_str ("$integration_method"), &
         var_str ("midpoint"), is_known = .true.)
    call dispatch_mci (mci, global, process_id)
    select type (mci)
    type is (mci_midpoint_t)
       call mci%write (u)
    end select

    call mci%final ()
    deallocate (mci)
    
    write (u, "(A)")
    write (u, "(A)")  "* Allocate MCI as vamp_t"
    write (u, "(A)")

    call var_list_set_string (global%var_list, &
         var_str ("$integration_method"), &
         var_str ("vamp"), is_known = .true.)
    call var_list_set_int (global%var_list, var_str ("threshold_calls"), &
         1, is_known = .true.)
    call var_list_set_int (global%var_list, var_str ("min_calls_per_channel"), &
         2, is_known = .true.)
    call var_list_set_int (global%var_list, var_str ("min_calls_per_bin"), &
         3, is_known = .true.)
    call var_list_set_int (global%var_list, var_str ("min_bins"), &
         4, is_known = .true.)
    call var_list_set_int (global%var_list, var_str ("max_bins"), &
         5, is_known = .true.)
    call var_list_set_log (global%var_list, var_str ("?stratified"), &
         .false., is_known = .true.)
    call var_list_set_log (global%var_list, var_str ("?use_vamp_equivalences"),&
         .false., is_known = .true.)
    call var_list_set_real (global%var_list, var_str ("channel_weights_power"),&
         4._default, is_known = .true.)

    call dispatch_mci (mci, global, process_id)
    select type (mci)
    type is (mci_vamp_t)
       call mci%write (u)
    end select

    call mci%final ()
    deallocate (mci)
    
    call global%final ()

    write (u, "(A)")
    write (u, "(A)")  "* Test output end: dispatch_3"
    
  end subroutine dispatch_3
  
  subroutine dispatch_4 (u)
    integer, intent(in) :: u
    type(rt_data_t), target :: global
    class(phs_config_t), allocatable :: phs
    
    write (u, "(A)")  "* Test output: dispatch_4"
    write (u, "(A)")  "*   Purpose: select phase-space configuration method"
    write (u, "(A)")

    call global%global_init ()

    write (u, "(A)")  "* Allocate PHS as phs_single_t"
    write (u, "(A)")

    call var_list_set_string (global%var_list, &
         var_str ("$phs_method"), &
         var_str ("single"), is_known = .true.)
    call dispatch_phs (phs, global)
    call phs%write (u)

    call phs%final ()
    deallocate (phs)
    
    write (u, "(A)")
    write (u, "(A)")  "* Allocate PHS as phs_wood_t"
    write (u, "(A)")

    call var_list_set_string (global%var_list, &
         var_str ("$phs_method"), &
         var_str ("wood"), is_known = .true.)
    call dispatch_phs (phs, global)
    call phs%write (u)

    call phs%final ()

    call global%final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: dispatch_4"
    
  end subroutine dispatch_4
  
  subroutine dispatch_5 (u)
    integer, intent(in) :: u
    type(rt_data_t), target :: global
    class(rng_t), allocatable :: rng
    
    write (u, "(A)")  "* Test output: dispatch_5"
    write (u, "(A)")  "*   Purpose: select random-number generator"
    write (u, "(A)")

    call global%global_init ()

    write (u, "(A)")  "* Allocate RNG as rng_test_t"
    write (u, "(A)")

    call var_list_set_string (global%var_list, &
         var_str ("$rng_method"), &
         var_str ("unit_test"), is_known = .true.)
    call dispatch_rng (rng, global)
    select type (rng)
    type is (rng_test_t)
       write (u, "(3x,A)")  "rng is rng_test_t"
    end select

    deallocate (rng)
    
    write (u, "(A)")
    write (u, "(A)")  "* Allocate RNG as rng_tao_t"
    write (u, "(A)")

    call var_list_set_string (global%var_list, &
         var_str ("$rng_method"), &
         var_str ("tao"), is_known = .true.)
    call dispatch_rng (rng, global)
    select type (rng)
    type is (rng_tao_t)
       write (u, "(3x,A)")  "rng is rng_tao_t"
    end select
    
    call global%final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: dispatch_5"
    
  end subroutine dispatch_5
  
  subroutine dispatch_6 (u)
    integer, intent(in) :: u
    type(rt_data_t), target :: global
    type(model_t), pointer :: model
    type(os_data_t) :: os_data
    type(process_constants_t) :: process_data
    class(phs_config_t), allocatable :: phs
    integer :: u_phs
    
    write (u, "(A)")  "* Test output: dispatch_6"
    write (u, "(A)")  "*   Purpose: select 'wood' phase-space &
         &for a test process"
    write (u, "(A)")  "*            and read phs configuration from file"
    write (u, "(A)")

    write (u, "(A)")  "* Initialize a process"
    write (u, "(A)")

    call global%global_init ()

    call os_data_init (os_data)
    call syntax_model_file_init ()
    call global%read_model (var_str ("Test"), var_str ("Test.mdl"))

    call syntax_phs_forest_init ()
    
    call init_test_process_data (var_str ("dispatch_6"), process_data)

    write (u, "(A)")  "* Write phase-space file"

    u_phs = free_unit ()
    open (u_phs, file = "dispatch_6.phs", action = "write", status = "replace")
    call write_test_phs_file (u_phs, var_str ("dispatch_6"))
    close (u_phs)

    write (u, "(A)")
    write (u, "(A)")  "* Allocate PHS as phs_wood_t"
    write (u, "(A)")

    call var_list_set_string (global%var_list, &
         var_str ("$phs_method"), &
         var_str ("wood"), is_known = .true.)
    call var_list_set_string (global%var_list, &
         var_str ("$phs_file"), &
         var_str ("dispatch_6.phs"), is_known = .true.)
    call dispatch_phs (phs, global)

    call phs%init (process_data, global%model)
    call phs%configure (sqrts = 1000._default)

    call phs%write (u)
    write (u, "(A)")
    select type (phs)
    type is (phs_wood_config_t)
       call phs%write_forest (u)
    end select

    call phs%final ()

    call global%final ()
    call syntax_model_file_final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: dispatch_6"
    
  end subroutine dispatch_6
  
  subroutine dispatch_7 (u)
    integer, intent(in) :: u
    type(rt_data_t), target :: global
    type(os_data_t) :: os_data
    type(string_t) :: prt, sf_method
    class(sf_data_t), allocatable :: data
    
    write (u, "(A)")  "* Test output: dispatch_7"
    write (u, "(A)")  "*   Purpose: select and configure &
         &structure function data"
    write (u, "(A)")

    call global%global_init ()
    
    call os_data_init (os_data)
    call syntax_model_file_init ()
    call global%read_model (var_str ("QCD"), var_str ("QCD.mdl"))
    
    call reset_interaction_counter ()
    call var_list_set_real (global%var_list, var_str ("sqrts"), &
         14000._default, is_known = .true.)
    prt = "p"
    call global%beam_structure%init ([prt, prt], [1])
    
    write (u, "(1x,A,L1)")  "Initialized = ", &
         pdf_builtin_status_is_initialized (global%pdf_builtin_status)
    write (u, "(A)")

    write (u, "(A)")  "* Allocate data as sf_pdf_builtin_t"
    write (u, "(A)")

    sf_method = "pdf_builtin"
    call dispatch_sf_data (data, sf_method, [1], global)
    call data%write (u)

    write (u, "(A)")
    write (u, "(1x,A,L1)")  "Initialized = ", &
         pdf_builtin_status_is_initialized (global%pdf_builtin_status)
    write (u, "(A)")
    write (u, "(1x,A,99(1x,I0))")  "PDG(out) = ", data%get_pdg_out ()

    deallocate (data)
    
    write (u, "(A)")
    write (u, "(A)")  "* Reset and allocate data for different PDF set"
    write (u, "(A)")

    call pdf_builtin_status_reset (global%pdf_builtin_status)
    
    write (u, "(1x,A,L1)")  "Initialized = ", &
         pdf_builtin_status_is_initialized (global%pdf_builtin_status)
    write (u, "(A)")

    call var_list_set_string (global%var_list, var_str ("$pdf_builtin_set"), &
         var_str ("CTEQ6M"), is_known = .true.)
    sf_method = "pdf_builtin"
    call dispatch_sf_data (data, sf_method, [1], global)
    call data%write (u)

    write (u, "(A)")
    write (u, "(1x,A,L1)")  "Initialized = ", &
         pdf_builtin_status_is_initialized (global%pdf_builtin_status)
    write (u, "(A)")
    write (u, "(1x,A,99(1x,I0))")  "PDG(out) = ", data%get_pdg_out ()

    deallocate (data)
    
    call global%final ()
    call syntax_model_file_final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: dispatch_7"
    
  end subroutine dispatch_7
  
  subroutine dispatch_8 (u)
    integer, intent(in) :: u
    type(rt_data_t), target :: global
    type(os_data_t) :: os_data
    type(flavor_t) :: flv
    type(sf_config_t), dimension(:), allocatable :: sf_config
    type(sf_channel_t), dimension(:), allocatable :: sf_channel
    type(string_t) :: sf_string
    integer :: i
    
    write (u, "(A)")  "* Test output: dispatch_8"
    write (u, "(A)")  "*   Purpose: configure a structure-function chain"
    write (u, "(A)")

    call global%global_init ()
    
    call os_data_init (os_data)
    call syntax_model_file_init ()
    call global%read_model (var_str ("QCD"), var_str ("QCD.mdl"))
    
    write (u, "(A)")  "* Allocate LHC beams with PDF builtin"
    write (u, "(A)")

    call flavor_init (flv, PROTON, global%model)

    call reset_interaction_counter ()
    call var_list_set_real (global%var_list, var_str ("sqrts"), &
         14000._default, is_known = .true.)
         
    call global%beam_structure%init (flavor_get_name ([flv, flv]), [1])
    call global%beam_structure%set_entry (1, 1, var_str ("pdf_builtin"))
    
    call dispatch_sf_config (sf_config, sf_channel, sf_string, global)

    do i = 1, size (sf_config)
       call sf_config(i)%write (u)
    end do
    write (u, "(1x,A)")  "Mapping configuration:"
    do i = 1, size (sf_channel)
       write (u, "(2x)", advance = "no")
       call sf_channel(i)%write (u)
    end do

    write (u, "(A)")
    write (u, "(A)")  "* Cleanup"

    call global%final ()
    call syntax_model_file_final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: dispatch_8"
    
  end subroutine dispatch_8
  
  subroutine dispatch_9 (u)
    integer, intent(in) :: u
    type(rt_data_t), target :: global
    class(eio_t), allocatable :: eio
    
    write (u, "(A)")  "* Test output: dispatch_9"
    write (u, "(A)")  "*   Purpose: allocate an event I/O (eio) stream"
    write (u, "(A)")

    call global%global_init ()
    
    write (u, "(A)")  "* Allocate as raw"
    write (u, "(A)")
    
    call dispatch_eio (eio, var_str ("raw"), global)

    call eio%write (u)

    call eio%final ()
    deallocate (eio)

    write (u, "(A)")
    write (u, "(A)")  "* Allocate as LHEF:"
    write (u, "(A)")
    
    call var_list_set_string (global%var_list, var_str ("$lhef_extension"), &
         var_str ("lhe_custom"), is_known = .true.)
    call dispatch_eio (eio, var_str ("lhef"), global)

    call eio%write (u)

    call eio%final ()
    deallocate (eio)

    write (u, "(A)")
    write (u, "(A)")  "* Allocate as HepMC:"
    write (u, "(A)")
    
    call dispatch_eio (eio, var_str ("hepmc"), global)

    call eio%write (u)

    call eio%final ()
    deallocate (eio)

    write (u, "(A)")
    write (u, "(A)")  "* Allocate as weight_stream"
    write (u, "(A)")
    
    call dispatch_eio (eio, var_str ("weight_stream"), global)

    call eio%write (u)

    write (u, "(A)")
    write (u, "(A)")  "* Cleanup"

    call eio%final ()
    call global%final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: dispatch_9"
    
  end subroutine dispatch_9
  
  subroutine dispatch_10 (u)
    integer, intent(in) :: u
    type(string_t), dimension(2) :: prt_in, prt_out
    type(rt_data_t), target :: global
    class(prc_core_def_t), allocatable :: core_def
    class(prc_core_t), allocatable :: core, saved_core
    type(var_list_t), pointer :: var_list
    
    write (u, "(A)")  "* Test output: dispatch_10"
    write (u, "(A)")  "*   Purpose: select process configuration method,"
    write (u, "(A)")  "             allocate process core,"
    write (u, "(A)")  "             temporarily reset parameters"
    write (u, "(A)")

    call syntax_model_file_init ()
    call global%global_init ()

    prt_in = [var_str ("a"), var_str ("b")]
    prt_out = [var_str ("c"), var_str ("d")]

    write (u, "(A)")  "* Allocate core as prc_omega_t"
    write (u, "(A)")

    call var_list_set_string (global%var_list, var_str ("$method"), &
         var_str ("omega"), is_known = .true.)
    call dispatch_core_def (core_def, prt_in, prt_out, global)

    call global%read_model (var_str ("Test"), var_str ("Test.mdl"))

    call dispatch_core (core, core_def, global%model)
    call core_def%allocate_driver (core%driver, var_str (""))

    select type (core)
    type is (prc_omega_t)
       call core%write (u)
    end select
    
    write (u, "(A)")
    write (u, "(A)")  "* Update core with modified model and helicity selection"
    write (u, "(A)")

    var_list => model_get_var_list_ptr (global%model)
    call var_list_set_real (var_list, var_str ("gy"), 2._default, &
         is_known = .true.)
    call model_parameters_update (global%model)

    call var_list_set_log (global%var_list, &
         var_str ("?helicity_selection_active"), &
         .true., is_known = .true.)
    call var_list_set_real (global%var_list, &
         var_str ("helicity_selection_threshold"), &
         2e10_default, is_known = .true.)
    call var_list_set_int (global%var_list, &
         var_str ("helicity_selection_cutoff"), &
         5, is_known = .true.)
    
    call dispatch_core_update (core, global%model, &
         global%get_helicity_selection (), &
         saved_core = saved_core)
    select type (core)
    type is (prc_omega_t)
       call core%write (u)
    end select
    
    write (u, "(A)")
    write (u, "(A)")  "* Restore core from save"
    write (u, "(A)")

    call dispatch_core_restore (core, saved_core)
    select type (core)
    type is (prc_omega_t)
       call core%write (u)
    end select
    
    call global%final ()
    call syntax_model_file_final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: dispatch_10"
    
  end subroutine dispatch_10
  
  subroutine dispatch_11 (u)
    integer, intent(in) :: u
    type(rt_data_t), target :: global
    type(qcd_t) :: qcd
    type(var_list_t), pointer :: model_vars
    
    write (u, "(A)")  "* Test output: dispatch_11"
    write (u, "(A)")  "*   Purpose: select QCD coupling formula"
    write (u, "(A)")

    call syntax_model_file_init ()
    call global%global_init ()
    call global%read_model (var_str ("SM"), var_str ("SM.mdl"))
    model_vars => model_get_var_list_ptr (global%model)

    write (u, "(A)")  "* Don't allocate: implies fixed alpha_s"
    write (u, "(A)")

    call var_list_set_log (global%var_list, var_str ("?alpha_s_is_fixed"), &
         .true., is_known = .true.)
    call dispatch_qcd (qcd, global)
    call qcd%write (u)

    write (u, "(A)")
    write (u, "(A)")  "* Allocate alpha_s as running (built-in)"
    write (u, "(A)")
    
    call var_list_set_log (global%var_list, var_str ("?alpha_s_is_fixed"), &
         .false., is_known = .true.)
    call var_list_set_log (global%var_list, var_str ("?alpha_s_from_mz"), &
         .true., is_known = .true.)
    call var_list_set_int &
         (global%var_list, var_str ("alpha_s_order"), 1, is_known = .true.)
    call var_list_set_real &
         (model_vars, var_str ("alphas"), 0.1234_default, &
          is_known=.true.)
    call var_list_set_real &
         (model_vars, var_str ("mZ"), 91.234_default, &
          is_known=.true.)
    call dispatch_qcd (qcd, global)
    call qcd%write (u)
    
    write (u, "(A)")
    write (u, "(A)")  "* Allocate alpha_s as running (built-in, Lambda defined)"
    write (u, "(A)")
    
    call var_list_set_log (global%var_list, var_str ("?alpha_s_is_fixed"), &
         .false., is_known = .true.)
    call var_list_set_log (global%var_list, var_str ("?alpha_s_from_mz"), &
         .false., is_known = .true.)
    call var_list_set_real &
         (global%var_list, var_str ("lambda_qcd"), 250.e-3_default, &
          is_known=.true.)
    call var_list_set_int &
         (global%var_list, var_str ("alpha_s_order"), 2, is_known = .true.)
    call var_list_set_int &
         (global%var_list, var_str ("alpha_s_nf"), 4, is_known = .true.)
    call dispatch_qcd (qcd, global)
    call qcd%write (u)
    
    write (u, "(A)")
    write (u, "(A)")  "* Allocate alpha_s as running (using builtin PDF set)"
    write (u, "(A)")
    
    call var_list_set_log (global%var_list, var_str ("?alpha_s_is_fixed"), &
         .false., is_known = .true.)
    call var_list_set_log (global%var_list, var_str ("?alpha_s_from_mz"), &
         .false., is_known = .true.)
    call var_list_set_log &
         (global%var_list, var_str ("?alpha_s_from_pdf_builtin"), &
         .true., is_known = .true.)
    call dispatch_qcd (qcd, global)
    call qcd%write (u)
    
    call global%final ()
    call syntax_model_file_final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: dispatch_11"
    
  end subroutine dispatch_11
  

end module dispatch
